/*
() Joker-tmn  - www.joker-tmn.narod.ru
  InfraHid
Copyright (C) 2007 Alex Badea <vamposdecampos@gmail.com>
http://vamposdecampos.googlepages.com/infrahid.html */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /*  sei() */
#include <util/delay.h>     /*  _delay_ms() */
#include <avr/eeprom.h>

#include <avr/pgmspace.h>   /*   usbdrv.h */
#include "usbdrv.h"
#include "oddebug.h"        /*        */

#define LED_BIT 0
#define LED_PORT PORTB
#define LED_PORT_DDR DDRB

#define led_on()  do { LED_PORT |= _BV(LED_BIT); } while (0)
#define led_off() do { LED_PORT &= ~_BV(LED_BIT); } while (0)
#define led_toggle()  do { LED_PORT ^= _BV(LED_BIT); } while (0)

/* ------------------------------------------------------------------------- */
//usb start
static const char hidReportDescriptor0[] PROGMEM = {
  /* partial keyboard */
  0x05, 0x01, /* Usage Page (Generic Desktop), */
  0x09, 0x06, /* Usage (Keyboard), */
  0xA1, 0x01, /* Collection (Application), */
  0x85, 0x01,   /* Report Id (1) */
  0x95, 0x04,   /* Report Count (4), */
  0x75, 0x08,   /* Report Size (8), */
  0x15, 0x00,   /* Logical Minimum (0), */
  0x25, 0x65,   /* Logical Maximum(101), */
  0x05, 0x07,   /* Usage Page (Key Codes), */
  0x19, 0x00,   /* Usage Minimum (0), */
  0x29, 0x65,   /* Usage Maximum (101), */
  0x81, 0x00,   /* Input (Data, Array),               ;Key arrays (4 bytes) */
  0xC0,   /* End Collection */

  /* mouse */
  0x05, 0x01, /* Usage Page (Generic Desktop), */
  0x09, 0x02, /* Usage (Mouse), */
  0xA1, 0x01, /* Collection (Application), */
  0x09, 0x01, /*   Usage (Pointer), */
  0xA1, 0x00, /*   Collection (Physical), */
  0x05, 0x09,   /* Usage Page (Buttons), */
  0x19, 0x01,   /* Usage Minimum (01), */
  0x29, 0x03,   /* Usage Maximun (03), */
  0x15, 0x00,   /* Logical Minimum (0), */
  0x25, 0x01,   /* Logical Maximum (1), */
  0x85, 0x02,   /* Report Id (2) */
  0x95, 0x03,   /* Report Count (3), */
  0x75, 0x01,   /* Report Size (1), */
  0x81, 0x02,   /* Input (Data, Variable, Absolute), ;3 button bits */
  0x95, 0x01,   /* Report Count (1), */
  0x75, 0x05,   /* Report Size (5), */
  0x81, 0x01,   /* Input (Constant),                 ;5 bit padding */
  0x05, 0x01,   /* Usage Page (Generic Desktop), */
  0x09, 0x30,   /* Usage (X), */
  0x09, 0x31,   /* Usage (Y), */
  0x15, 0x81,   /* Logical Minimum (-127), */
  0x25, 0x7F,   /* Logical Maximum (127), */
  0x75, 0x08,   /* Report Size (8), */
  0x95, 0x02,   /* Report Count (2), */
  0x81, 0x06,   /* Input (Data, Variable, Relative), ;2 position bytes (X & Y) */
  0xC0,   /*   End Collection, */
  0xC0,   /* End Collection */
};


static const char configDescriptor[] PROGMEM = {  /* USB configuration descriptor */
  9,      /* sizeof(usbDescriptorConfiguration): length of descriptor in bytes */
  USBDESCR_CONFIG,  /* descriptor type */
  9 + 1 * (9 + 9 + 7), 0,
  /* total length of data returned (including inlined descriptors) */
  1,      /* number of interfaces in this configuration */
  1,      /* index of this configuration */
  0,      /* configuration name string index */
  USBATTR_BUSPOWER, /* attributes */
  USB_CFG_MAX_BUS_POWER / 2,  /* max USB current in 2mA units */

/* --- interface 0 --- */
  9,      /* sizeof(usbDescrInterface): length of descriptor in bytes */
  USBDESCR_INTERFACE, /* descriptor type */
  0,      /* index of this interface */
  0,      /* alternate setting for this interface */
  1,      /* endpoints excl 0: number of endpoint descriptors to follow */
  3,      /* 3=HID class */
  0,      /* subclass: 0=none, 1=boot */
  0,      /* protocol: 0=none, 1=keyboard, 2=mouse */
  0,      /* string index for interface */

  /* HID descriptor */
  9,      /* sizeof(usbDescrHID): length of descriptor in bytes */
  USBDESCR_HID,   /* descriptor type: HID */
  0x01, 0x01,   /* BCD representation of HID version */
  0x00,     /* target country code */
  0x01,     /* number of HID Report (or other HID class) Descriptor infos to follow */
  0x22,     /* descriptor type: report */
  sizeof(hidReportDescriptor0), 0,  /* total length of report descriptor */

  /* endpoint descriptor for endpoint 1 */
  7,      /* sizeof(usbDescrEndpoint) */
  USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
  0x81,     /* IN endpoint number 1 */
  0x03,     /* attrib: Interrupt endpoint */
  8, 0,     /* maximum packet size */
  USB_CFG_INTR_POLL_INTERVAL, /* in ms */
};

uchar usbFunctionDescriptor(usbRequest_t * rq)
{
  uchar *p = 0, len = 0;
  DBG1(0xa8, &rq->wValue.bytes[1], 1);

#if 0 /* default descriptor from driver */
  if (rq->wValue.bytes[1] == USBDESCR_DEVICE) {
    p = (uchar *) deviceDescriptor;
    len = sizeof(deviceDescriptor);
  } else
#endif
  if (rq->wValue.bytes[1] == USBDESCR_CONFIG) {
    p = (uchar *) configDescriptor;
    len = sizeof(configDescriptor);
  } else if (rq->wValue.bytes[1] == USBDESCR_HID) {
    p = (uchar *) (configDescriptor + 18);
    len = 9;
  } else if (rq->wValue.bytes[1] == USBDESCR_HID_REPORT) {
    p = (uchar *) hidReportDescriptor0;
    len = sizeof(hidReportDescriptor0);
  }
  usbMsgPtr = p;
  return len;
}

static uchar inputBuffer_kb_msg[5] = {1,0,0,0,0,};   /* kbd */
static uchar inputBuffer_mouse_msg[4] = { 2, 0,0,0, };   /* mouse */
static uchar idleRate = 0;        /* in 4 ms units */
static uchar tx_kbd = 0;
static uchar tx_mouse = 0;
static uchar dev_mode = 0;//  -  !

uchar usbFunctionSetup(uchar data[8])
{
  usbRequest_t *rq = (void *) data;

  DBG1(0xa0, data, 8);

  if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {  /* class request type */
    if (rq->bRequest == USBRQ_HID_GET_REPORT) {
      /* wValue: ReportType (highbyte), ReportID (lowbyte) */
      if (rq->wValue.bytes[0] == 1) 
        {
        usbMsgPtr = inputBuffer_kb_msg;
        return sizeof(inputBuffer_kb_msg);
        }
      return 0;
    } else if (rq->bRequest == USBRQ_HID_SET_REPORT) {
      /* wValue: ReportType (highbyte), ReportID (lowbyte) */
      /* we have no output/feature reports */
      return 0;
    } else if (rq->bRequest == USBRQ_HID_GET_IDLE) {
      usbMsgPtr = &idleRate;
      return 1;
    } else if (rq->bRequest == USBRQ_HID_SET_IDLE) {
      idleRate = rq->wValue.bytes[1];
    }
    return 0;
  }

  return 0;
}
//usb end
/* ------------------------------------------------------------------------- */

/*    IR */
static const struct {
  unsigned short
    header_hi, header_lo, repeat_lo,
    bit_hi, one_lo, zero_lo;
} irconf = {
/* original values (in microseconds)
  9062, 4442, 2193,
  616, 1633, 508,
*/
/* values in 5.33us units: */
  1699, 833, 411,
  115, 306, 95,
};

static struct code_entry {unsigned long remote_code; uchar hid_code; uchar cmd_code;} code_map[] = 
  {
    { 0xf50a6c93, 0x00 , 0x01}, /* kb_or_mouse_mode_change - phone */
    { 0xf50a4cb3, 0x53 , 0x02}, /* space - green phone */
    { 0xf50acc33, 0x2c, 0x03}, /* space  - red phone*/

    { 0xb54ac23d, 0x5c , 0x10}, /* left */
    { 0xb54a42bd, 0x5e  , 0x11}, /* Right */
    { 0xb54a02fd, 0x5a , 0x12}, /* Down */
    { 0xb54a827D, 0x60 , 0x13}  /* Up */
  }; 
#define NUM_CODES (sizeof(code_map) / sizeof(struct code_entry))

#define IR_TIMEOUT_CONST  8 /* in 21.8ms intervals */

/* ir_state values */
#define IR_IDLE   0
#define IR_HEADER 1
#define IR_DATA_HI  2
#define IR_DATA_LO  3
#define IR_DISCARD  4

#define IR_ACCEL_DEF  2
#define IR_ACCEL_INC_LO 1
#define IR_ACCEL_INC_HI 2
#define IR_ACCEL_MAX  50

static uchar ir_state = IR_IDLE, ir_bits = 0;
static unsigned long ir_data = 0, prev_ir_data = 0;
static uchar ir_timeout = 0;
static uchar ir_accel = IR_ACCEL_DEF;

#define IR_BUFLEN 8 /* this _must_ be an even value */
static unsigned short ir_buf[IR_BUFLEN];
static uchar ir_head = 0;
static uchar ir_tail = 0;

SIGNAL(SIG_INTERRUPT1) //  IR
{
  ir_buf[ir_head] = TCNT1;  /* store pulse length */
  ir_head = (ir_head + 1) % IR_BUFLEN;

  MCUCR ^= _BV(ISC10);    /* toggle rising/falling edge */
  TCNT1 = 0;      /* clear counter */
}

#define ABS(x)  ((x) < 0 ? -(x) : (x))

static uchar ir_check(const unsigned short ref, unsigned short value)
{
  short delta = ABS((short) value - (short) ref);
  if (delta > 100)    /* aeps = 100us */
    return 0;
  if (delta > ref / 3)    /* eps = 30%*/
    return 0;
  return 1;
}

static uchar ir_reset(void)
{
  ir_state = IR_IDLE;
  ir_bits = 0;
  return 0;
}

static void hid_clear(void)
{
  inputBuffer_mouse_msg[1] = 0;  /* mouse buttons */
  inputBuffer_mouse_msg[2] = 0;  /* X axis delta */
  inputBuffer_mouse_msg[3] = 0;  /* Y axis delta */

  inputBuffer_kb_msg[1] = 0;  /* key */
}

static void ir_input(void)
{    
    uchar k;

/*  if (ir_data == 0xa857)  // "scan" button 
    reset();  // XXX HACK
*/
    if (!usbInterruptIsReady()) 
      {
      led_on();
      return;   // still sending interrupt packet
      }
    hid_clear();

    if ((prev_ir_data==ir_data)||(ir_data==0xB54AE916))//0xB54AE916 - 
      {
      if (ir_accel < IR_ACCEL_MAX / 2)
          ir_accel += IR_ACCEL_INC_LO;
      else 
          if (ir_accel < IR_ACCEL_MAX)
              ir_accel += IR_ACCEL_INC_HI;
      } 
    else 
      {
      ir_accel = IR_ACCEL_DEF;
      }

    for (k = 0; k < NUM_CODES; k++) 
      {
      if (ir_data != code_map[k].remote_code)
        continue;

      if (code_map[k].cmd_code == 0x01)
         {
           if ((prev_ir_data==ir_data)||(ir_data==0xB54AE916))//0xB54AE916 - 
                {break;}
           else
                {               
                if (dev_mode==1)
                   {dev_mode=0;}
		else
 		   {dev_mode=1;}		
                }         
           break;
         }
      
      if(usbConfiguration!=0) /*   */
             {
             if (dev_mode!=0)
               {
	       {inputBuffer_kb_msg[1] = code_map[k].hid_code;}
               tx_kbd = 1;  // send kbd event 
               usbSetInterrupt(inputBuffer_kb_msg, sizeof(inputBuffer_kb_msg));
               }
             else
               {
               if (code_map[k].cmd_code == 0x10)
                  inputBuffer_mouse_msg[2] = -ir_accel; /* left */
               else if (code_map[k].cmd_code == 0x11)
                  inputBuffer_mouse_msg[2] = ir_accel; /* right */
               else if (code_map[k].cmd_code == 0x12)
                  inputBuffer_mouse_msg[3] = ir_accel; /* down */
               else if (code_map[k].cmd_code == 0x13)
                  inputBuffer_mouse_msg[3] = -ir_accel; /* up */
               else if (code_map[k].cmd_code == 0x02)
                  inputBuffer_mouse_msg[1] |= 2 ; /* rb click */
               else if (code_map[k].cmd_code == 0x03)
                  inputBuffer_mouse_msg[1] |= 1; /* lb click */
               tx_mouse = 1;  // send kbd event 
               usbSetInterrupt(inputBuffer_mouse_msg, sizeof(inputBuffer_mouse_msg));
               }
             }
      break;      
      }
    prev_ir_data=ir_data;
    ir_timeout = IR_TIMEOUT_CONST;
    led_toggle();
}

static uchar ir_decode(uchar level, unsigned short time)
{
  /* as pointed out by other people as well, if () cascades are better
   * than switch () since they operate on chars instead of ints. */
  if (ir_state == IR_DISCARD) {    
    /* counter overflowed */
    return ir_reset();
  } else if (ir_state == IR_IDLE) {
    if (level && ir_check(irconf.header_hi, time)) {
      ir_state = IR_HEADER;
    } else {
      return ir_reset();
    }
  } else if (ir_state == IR_HEADER) {
    if (!level && ir_check(irconf.repeat_lo, time)) {
      DBG1(0x99, (uchar *) &ir_data, 2);
      ir_input();
      return ir_reset();
    } else if (!level && ir_check(irconf.header_lo, time)) {
      ir_state = IR_DATA_HI;
      ir_data = 0;
    } else {
      return ir_reset();
    }
  } else if (ir_state == IR_DATA_HI) {
    if (level && ir_check(irconf.bit_hi, time)) {
      ir_state = IR_DATA_LO;
    } else {
      return ir_reset();
    }
  } else if (ir_state == IR_DATA_LO) {
    ir_bits++;
    ir_data <<= 1;
    if (!level && ir_check(irconf.one_lo, time)) {
      ir_data |= 1;
    } else if (!level && ir_check(irconf.zero_lo, time)) {
      /* nothing */
    } else {
      return ir_reset();
    }
    ir_state = IR_DATA_HI;

    if (ir_bits == 32) 
          {
          DBG1(0x88, (uchar *) &ir_data, 2);
          ir_input();
          ir_reset();
          }
  }
  return 1;
}

static void ir_queue_poll(void)
{
  /* if any pulses were received, decode them */
  while (ir_head != ir_tail) {
    unsigned short timing = ir_buf[ir_tail];
    ir_decode(ir_tail & 1, timing);
    ir_tail = (ir_tail + 1) % IR_BUFLEN;
  }
}

/* this may well be SIGNAL(SIG_OVERFLOW0), but we want it called at well-defined times */
static void idle_timer(void)
{   
  static uchar idleCounter = 0;
  /* idle reports, if requested */
  if (idleRate) {
    if (idleCounter > 4) {
      idleCounter -= 5;
    } else {
      idleCounter = idleRate;
      tx_kbd = 1;
      tx_mouse = 1;
    }
  }
      
  /* process button releases */
  if (ir_timeout) {
    if (!--ir_timeout) {
      hid_clear();
      tx_kbd = 1;
      tx_mouse = 1;
      led_on();
    }
  }
}

static void send_packets(void)
{
  /* send pending packets */
  if (usbInterruptIsReady()) 
   {
    if (tx_kbd) 
      {
      usbSetInterrupt(inputBuffer_kb_msg, sizeof(inputBuffer_kb_msg));  /* send kbd event */
      tx_kbd = 0;
      }
    else 
     if (tx_mouse)  
      {
      usbSetInterrupt(inputBuffer_mouse_msg, sizeof(inputBuffer_mouse_msg));  /* send mouse event */
      tx_mouse = 0;
      }
    hid_clear();
   }
}

/* ------------------------------------------------------------------------- */
int main(void)
{
    uchar i=0;
    
    wdt_enable(WDTO_1S);
    /*      watchdog,   .   
     *   watchdog (/, )   RESET! */

    // Port B initialization
    PORTB=0xff;
    DDRB=0x00;

    // Port C initialization
    PORTC=0xff;
    DDRC=0x00;

    /*    */    
    PORTD |= _BV(1);
    DDRD |= _BV(1);

    // Analog Comparator initialization
    // Analog Comparator: Off
    // Analog Comparator Input Capture by Timer/Counter 1: Off
    ACSR=0x80;
    SFIOR=0x00;

    // ADC initialization
    ADMUX=0xff;
    ADCSRA=0x84;

    //    INT1
    GICR |= _BV(INT1);  /* enable INT1 */
    MCUCR = (MCUCR & ~(_BV(ISC11) | _BV(ISC10))) | _BV(ISC11);  /* set INT1 on falling edge */
    TCCR0 = 5;    /* prescaler = 1/1024 -> overflow at 21.8 ms */
    TCCR1B = 3;   /* prescaler = 1/64 -> 1 count = 5.33333 us */


    DBG1(0x00, 0, 0);       /*  :   main */
    /*  RESET:          (pull-up).
     *  ,    D+  D-.  ,      .
     */

    odDebugInit();
    usbInit();
    usbDeviceDisconnect();  /*   ,  ,   ! */
    i = 0;
    while(--i){             /*  USB    > 250  */
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();

    sei();
    DBG1(0x01, 0, 0);       /*  :    main */

    LED_PORT_DDR = _BV(LED_BIT);  /* LED    */
    led_on(); //LED   IR

    for(;;)
    {                /*   main */
        DBG1(0x02, 0, 0);   /*  :   main */
        wdt_reset();
        usbPoll();

        ir_queue_poll();

        /* on timer overflow (~22ms) */
        if (TIFR & _BV(TOV0)) 
          { TIFR = _BV(TOV0);
            idle_timer(); }

        if(usbConfiguration!=0) 
          {
           PORTD &= ~_BV(1);/*   0, ..     +5!!!*/
           /*   */
           send_packets();//              
          }
        else
          {
           PORTD |= _BV(1);/*  */
          }
    }
    return 0;
}
/* ------------------------------------------------------------------------- */
